/* ----------------------------------------------------------------------- *
 *   
 *   Copyright 2003 H. Peter Anvin - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 53 Temple Place Ste 330,
 *   Bostom MA 02111-1307, USA; either version 2 of the License, or
 *   (at your option) any later version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

module cfstream(
		clk,		// 2-20 MHz
		reset_n,
		
		cf_a,
		cf_d,
		cf_bsy_n,
		cf_wait_n,
		cf_ce1_n,
		cf_ce2_n,
		cf_oe_n,
		cf_we_n,
		cf_reg_n,

		q,
		data_ready_n,
		rdreq,

		bsy_time
		);

   input         clk;
   input	 reset_n;
   output [10:0] cf_a;
   inout  [15:0] cf_d;
   input 	 cf_bsy_n;
   input 	 cf_wait_n;
   output 	 cf_ce1_n, cf_ce2_n;
   output 	 cf_oe_n;
   output 	 cf_we_n;
   output 	 cf_reg_n;

   output [15:0] q;
   output 	 data_ready_n;
   input 	 rdreq;

   output [7:0]  bsy_time;

   reg [10:0] 	 cf_a = 0;
   reg           cf_oe_n = 1;
   reg           cf_we_n = 1;

   reg [27:0] 	 lba = 0;
   reg [7:0] 	 wordix;	// 256 words @ 16 bits/sector

   reg [15:0] 	 cf_data;
   reg 	         wrreq;
   wire 	 fifo_full;
   wire 	 fifo_almost_full;
   
   reg		 cf_wait_q;
   
   // Megafunction FIFO
   cffifo cffifo_inst (
		       .data ( cf_data ),
		       .wrreq ( wrreq ),
		       .rdreq ( rdreq ),
		       .clock ( clk ),
		       .q ( q ),
		       .full ( fifo_full ),
		       .empty ( data_ready_n ),
		       .almost_full ( fifo_almost_full ) // Less than 128 words free
		       );
   
`define FSM_IDLE	4'h0
`define FSM_SEND_23	4'h1
`define FSM_WAIT_23     4'h2
`define FSM_SEND_45     4'h3
`define FSM_WAIT_45     4'h4
`define FSM_SEND_67     4'h5
`define FSM_WAIT_67     4'h6
`define FSM_WAIT_BSY    4'h7
`define FSM_WAIT_RDY    4'h8
`define FSM_GET_DATA	4'h9
`define FSM_WAIT_DATA	4'hA

   reg [3:0] 	 state;

   reg [15:0] 	 out_d;
   reg           drive_d;
   assign        cf_d    = drive_d ? out_d : 16'bz;

   assign 	 cf_reg_n = 1;    // Common memory
   assign 	 cf_ce1_n = 0;	  // 16-bit Common Memory
   assign 	 cf_ce2_n = 0;	  // 16-bit Common Memory

   reg [23:0] 	 bsy_ctr;
   reg [7:0] 	 bsy_time_q = 0;
   assign 	 bsy_time = bsy_time_q;
   
   always @(negedge reset_n or posedge clk)
     if ( ~reset_n )
       begin
	  cf_a       <= 0;
	  lba        <= 0;
	  state      <= `FSM_IDLE;
	  drive_d    <= 0;
	  cf_we_n    <= 1;
	  cf_oe_n    <= 1;
	  bsy_time_q <= 0;
       end
     else
       begin
	  wrreq <= 0;
	  cf_wait_q <= cf_wait_n;
	  
	  case (state)
	    `FSM_IDLE:
	      begin
		 cf_a        	<= 11'h002;
		 out_d 		<= 16'bx;
		 drive_d   	<= 0;
		 cf_we_n	<= 1;
		 cf_oe_n      	<= 1;
		 
		 if ( !fifo_almost_full )
		   state		<= `FSM_SEND_23;
	      end
	    
	    `FSM_SEND_23:
	      begin
		 out_d[7:0]	<= 8'h01;
		 out_d[15:8] 	<= lba[7:0];
		 drive_d     	<= 1;
		 cf_we_n      	<= 0;
		 cf_oe_n      	<= 1;
		 cf_wait_q      <= 0;
		 state       	<= `FSM_WAIT_23;
	      end
	    
	    `FSM_WAIT_23:
	      begin
		 if ( cf_wait_q )
		   begin
		      cf_a      <= 11'h004;
		      state	<= `FSM_SEND_45;
		      cf_we_n	<= 1;
		   end
	    end
	    
	    `FSM_SEND_45:
	      begin
		 out_d       	<= lba[23:8];
		 drive_d     	<= 1;
		 cf_we_n      	<= 0;
		 cf_oe_n      	<= 1;
		 cf_wait_q      <= 0;
		 state       	<= `FSM_WAIT_45;
	      end
	    
	    `FSM_WAIT_45:
	      begin
		 if ( cf_wait_q )
		   begin
		      cf_a	<= 11'h006;
		      state	<= `FSM_SEND_67;
		      cf_we_n	<= 1;
		   end
	      end
	    
	    `FSM_SEND_67:
	      begin
		 out_d[3:0]  	<= lba[27:24];
		 out_d[7:4]  	<= 4'hE;
		 out_d[15:8] 	<= 8'h20;
		 drive_d     	<= 1;
		 cf_we_n      	<= 0;
		 cf_oe_n     	<= 1;
		 cf_wait_q      <= 0;
		 state       	<= `FSM_WAIT_67;
		 
		 lba		<= lba + 1;
	      end
	    
	    `FSM_WAIT_67:
	      begin
		 bsy_ctr	<= 0;
		 
		 if ( cf_wait_q )
		   begin
		      cf_a	<= 11'h000;
		      state	<= cf_bsy_n ? `FSM_WAIT_BSY : `FSM_WAIT_RDY;
		      cf_we_n	<= 1;
		      drive_d	<= 0;
		   end
		 wordix     	<= 0;
	      end
	    
	    `FSM_WAIT_BSY:
	      begin
		 bsy_ctr	<= bsy_ctr + 1;
		 if ( !cf_bsy_n )
	      	   state 	<= `FSM_WAIT_RDY;
	      end
	  
	  `FSM_WAIT_RDY:
	    begin
	       bsy_ctr		<= bsy_ctr + 1;
	       if ( cf_bsy_n )
	      	 state 		<= `FSM_GET_DATA;
	    end
	  
	    `FSM_GET_DATA:
	      begin
		 if ( bsy_time_q < bsy_ctr[23:16] )
		   bsy_time_q	<= bsy_ctr[23:16];
		 
		 cf_a      	<= 11'h000;
		 drive_d   	<= 0;
		 cf_we_n    	<= 1;
		 cf_oe_n    	<= 0;
		 cf_wait_q      <= 0;
		 state     	<= `FSM_WAIT_DATA;
	      end

	    `FSM_WAIT_DATA:
	      begin
		 if ( cf_wait_q )
		   begin
		      cf_oe_n	<= 1;
		      cf_data   <= cf_d;
		      wrreq 	<= 1;
		      state	<= ( &wordix ) ? `FSM_IDLE : `FSM_GET_DATA;
		      wordix 	<= wordix + 1;
		   end
	      end
	    
	    default:
	      begin
		 cf_a   	<= 11'bx;
		 out_d		<= 16'bx;
		 drive_d  	<= 1'bx;
		 cf_we_n 	<= 1'bx;
		 cf_oe_n 	<= 1'bx;
		 state  	<= 4'bx;
	    end
	  endcase
       end // else: !if( ~reset_n )
endmodule
