/*
 * sound.v
 *
 * Circuit to generate PDM sound using an Altera Nios board with an
 * AleaRep Lancelot sound add-on.
 *
 * This module should be fed with a 200 MHz clock from the PLL.
 * 
 * This is a delta-sigma DAC, assuming an external low pass filter.
 */

`define BITS	16
`define MSB	(`BITS-1)
`define DBITS	(`BITS+1)

module sound(
	clk,
	data_l,
	data_r,
	audio_l,
	audio_r,
	audio_clk
	);

   input	clk;
   input [`MSB:0] data_l;
   input [`MSB:0] data_r;
   output	audio_l;
   output	audio_r;
   output       audio_clk;

   reg [`MSB:0] 	data_latch_l;
   reg [`MSB:0] 	data_latch_r;

   wire [`DBITS:0] 	delta_add_l;
   wire [`DBITS:0] 	delta_add_r;

   wire [`DBITS:0] 	sigma_add_l;
   wire [`DBITS:0] 	sigma_add_r;

   reg [`DBITS:0] 	sigma_latch_l;
   reg [`DBITS:0] 	sigma_latch_r;

   wire [`DBITS:0] 	delta_l;
   wire [`DBITS:0] 	delta_r;

   reg 		out_latch_l;
   reg 		out_latch_r;

   reg [12:0] 	pulse_ctr;
   
   // This gives about 12-bit sound at 44100 Hz.  A high sample rate will
   // help avoid actually making the sampling frequency audible.
   // The low-pass filter on the Lancelot has its knee at 10 kHz, and
   // (being a first-order filter) drops off at 3 dB/octave.

   // Note: the input is expected unsigned.  For signed input, invert
   // bit `MSB.
   
   assign 	delta_add_l = { 2'b0, data_latch_l } + delta_l;
   assign 	sigma_add_l = delta_add_l + sigma_latch_l;
   assign 	delta_l[`DBITS]     = sigma_latch_l[`DBITS];
   assign 	delta_l[`DBITS-1]   = sigma_latch_l[`DBITS];
   assign 	delta_l[`DBITS-2:0] = 0;

   assign 	delta_add_r = { 2'b0, data_latch_r } + delta_r;
   assign 	sigma_add_r = delta_add_r + sigma_latch_r;
   assign 	delta_r[`DBITS]     = sigma_latch_r[`DBITS];
   assign 	delta_r[`DBITS-1]   = sigma_latch_r[`DBITS];
   assign 	delta_r[`DBITS-2:0] = 0;
   
   always @(posedge clk)
     begin
	if ( pulse_ctr == 0 )
	  begin
	     pulse_ctr <= 4535-1;      // 44100 Hz sample rate
	     data_latch_l <= data_l;
	     data_latch_r <= data_r;
	  end
	else
	  begin
	     pulse_ctr <= pulse_ctr-1;
	  end

	out_latch_l <= sigma_latch_l[`DBITS];
	sigma_latch_l <= sigma_add_l;
	
	out_latch_r <= sigma_latch_r[`DBITS];
	sigma_latch_r <= sigma_add_r;
     end

   // Definitely *not* a 50% duty cycle; goes high immediately after
   // the data item has been latched
   assign audio_clk = pulse_ctr[12];

   assign audio_l = out_latch_l;
   assign audio_r = out_latch_r;
   
endmodule // sound
